/**
 *  UNWAR
 *  UNpacker for Werkkzeug ARchives.
 *  
 *  (c) 2011, Sauron.
 */

#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>

#include "unwar.h"

// global variables
byte *front;
uint32 window;
sint32 range;
uint32 dictionary[0x400];

// Allocate memory and exit() if cannot allocate
void *xmalloc(size_t size) {
    void *data=malloc(size);
    if (!data) {
        fprintf(stderr,"Cannot allocate %d bytes memory!\n",size);
        exit(1);
    };
    return data;
}

uint32 decodeBit(uint32 shift, uint32 index) {
    sint32 data=(sint32)dictionary[index];
    sint32 bound;
    uint32 retval;
    
    bound=(sint32)((uint32)range>>11);
    bound=bound*data;

    if (window<(uint32)bound) {
        range=bound;
        data=dictionary[index];
        data=((0x800-data)>>(shift&0xff))+data;
        dictionary[index]=data;
        retval=0;
    }
    else {
        window-=bound;
        range-=bound;
        data=data-(data>>(shift&0xff));
        dictionary[index]=data;
        retval=1;
    };
    
    // if end of byte reached, move window right
    if ((uint32)range<0x01000000) {
        window=(window<<8)|((uint32)(*front));
        ++front;
        range<<=8;
    };
    
    return retval;
}

uint32 decodeValue(uint32 index) {
    uint32 retval=1;
    byte ctx2=1;
    uint32 ctx;
    
    do {
        ctx=(byte)(ctx2*2+decodeBit(5,index+ctx2));
        retval=decodeBit(5,ctx+index)+2*retval;
        ctx2=(retval&1)+(byte)(ctx*2);
    } while ((ctx2&2)!=0);
    
    return retval;
}

byte *_decompress(byte *data, byte *dest) {
    uint32 i;
    uint32 offset_last=0;
    uint32 tmp=0;
    uint32 mode=0;
    uint32 lastMode;
    uint32 retval;
    
    //initialization
    front=data+4;
    window=((uint32)data[0]<<24)|((uint32)data[1]<<16)|((uint32)data[2]<<8)|((uint32)data[3]<<0);
    range=-1;
    for (i=0; i<=0x323; i++)
        dictionary[i]=0x400;

    //decompress
    for (;;) {
        if (mode==1) {
            // repeatable fragment
            uint32 count=0;
            uint32 offset=0;
            if ((lastMode!=0)||(decodeBit(5,2)==0)) {
                offset=decodeValue(0x123);
                if (offset==0)
                    break; // end of file
                
                offset-=2;
                tmp=(offset?16:0)+3;
                uint32 accumulator=1;
                do {
                    retval=decodeBit(5,accumulator+tmp);
                    accumulator=retval+2*accumulator;
                } while (accumulator<16);
                offset=(offset<<4)+accumulator-15;
                if (offset>=0x800)
                    count=1;
                if (offset>=0x60)
                    count++;
            }
            else
                offset=offset_last;
            
            offset_last=offset;
            count+=decodeValue(0x223);
            if (count>0) {
                byte *dest2=dest-offset;
                do {
                    *dest=*dest2;
                    dest++;
                    dest2++;
                    count--;
                } while ((sint32)count>0);
                lastMode=1;
            };
        }
        else if (mode==0) {
            // literal
            uint32 accumulator=1;
            do {
                retval=decodeBit(4,accumulator+0x23);
                accumulator=retval+2*accumulator;
            } while (accumulator<0x100);

            *dest=(byte)accumulator;

            dest++;
            lastMode=0;
        };
        
        // what is the next bit?
        mode=decodeBit(5,lastMode);
    };

    return dest;
}

byte *decompress(byte *data, uint32 size) {
    byte *dest=(byte *)xmalloc(size+1);
    if (!data)
        return NULL;
    else {
        _decompress(data,dest);
        return dest;
    };
}

void *file_get_contents(const char *filename) {
    void *data;
    int length;

    FILE *f=fopen(filename,"r");
    if (f==NULL)
        return NULL;
    fseek(f,0,SEEK_END);
    length=ftell(f);
    fseek(f,0,SEEK_SET);
    data=xmalloc(length);
    fread(data,length,1,f);
    fclose(f);
    
    return data;
}

void *read_fragment(int fd, int offset, int length) {
    lseek(fd,offset,SEEK_SET);
    void *data=malloc(length+1);
    if (!data)
        return NULL;
    read(fd,data,length);
    return data;
}

int file_set_contents(const char *filename, void *buf, int size) {
    FILE *f=fopen(filename,"w");
    if (f==NULL)
        return 0;
    fwrite(buf,size,1,f);
    fclose(f);
    return 1;
}

char **read_names(int fd, struct wz4_descriptor *files, int nrfiles, int data_offset) {
    int min_offset=data_offset;
    int i;
    int j;
    char *names_raw;
    int name_offset;
    char name[256];
    char **result;
    
    for (i=0; i<nrfiles; i++) {
        if (min_offset>files[i].nameOffset)
            min_offset=files[i].nameOffset;
    };
    
    names_raw=(char *)read_fragment(fd,min_offset,data_offset-min_offset);
    
    result=(char **)xmalloc(nrfiles*sizeof(char *));
    
    for (i=0; i<nrfiles; i++) {
        name_offset=files[i].nameOffset-min_offset;
        for (j=0; j<256; j++) {
            name[j]=names_raw[name_offset+j*2];
            if (name[j]=='\0')
                break;
        };
        result[i]=xmalloc(j+1);
        strcpy(result[i],name);
    };
    
    // free memory
    free(names_raw);
    
    return result;
}

int unpack_archive(const char *filename, int mode, const char *dir) {
    int fd=open(filename,O_RDONLY);
    struct wz4_descriptor *files;
    int i;
    char **names;
    byte *fragment;
    byte *fragment_unpacked;
    
    // metadata
    struct wz4_pak_header header;
    read(fd,&header,sizeof header);
    if (header.magic!=PAK_MAGIC) {
        fprintf(stderr,"It is not WZ4 package! Magic=0x%08X\n",header.magic);
        return 3;
    };
    
    if (header.nrfiles==0) {
        fprintf(stderr,"No files found!\n");
        return 0;
    };
    
    // descriptors
    files=(struct wz4_descriptor *)xmalloc(header.nrfiles*sizeof(struct wz4_descriptor));
    read(fd,files,sizeof(struct wz4_descriptor)*header.nrfiles);
    
    // names
    names=read_names(fd,files,header.nrfiles,header.dataOffset);
    
    if (mode==0) {
        // output filenames
        for (i=0; i<header.nrfiles; i++)
            printf("[%d]: %s\n",i,names[i]);
    }
    else {
        if (dir!=NULL) {
            if (chdir(dir)<0) {
                if (mkdir(dir,0770)<0) {
                    fprintf(stderr,"Cannot create directory %s\n",dir);
                    return 4;
                };
                chdir(dir);
            };
        };
        
        // unpack archive
        for (i=0; i<header.nrfiles; i++) {
            fragment=(byte *)read_fragment(fd,files[i].fileOffset,files[i].packedSize);
            if (files[i].size!=files[i].packedSize) {
                //printf("packed\n");
                fragment_unpacked=decompress(fragment,files[i].size);
                free(fragment);
            }
            else {
                //printf("stored\n");
                fragment_unpacked=fragment;
            };
            
            file_set_contents(names[i],fragment_unpacked,files[i].size);
            free(fragment_unpacked);
        };
    };
    
    // free memory
    free(files);
    free(names);
    
    close(fd);
    
    return 0;
}

int main(int argc, char **argv) {
    int action=1;
    
    if (argc<2) {
        fprintf(stderr,"Too few arguments!\n");
        return 2;
    };
    
    char *dir=NULL;
    if (argc>2) {
        dir=argv[2];
    };
    
    return unpack_archive(argv[1],1,dir);
}

